def render_class_frame(class_body):
    return f"""from commondao.mapper import Mapper
from datetime import datetime, date
from lessweb.bridge import service


@service
class CommonDao(Mapper):
{class_body}
"""


def render_class_methods(table_name, entity_name, col_items, unique_keys):
    """
    col_items: 形如dict[name, (py_type, comment)]
    unique_keys: 形如list[list[str]]
    """
    indent = '    '
    kwargs = ''.join(f'{indent * 2}{name}: {py_type} = None,  # {comment}\n'
                     for name, (py_type, comment) in col_items.items())
    data_rows = ''.join(f"{indent * 3}'{name}': {name},\n"
                        for name in col_items.keys())
    by_key_methods = ''
    for unique_key in unique_keys:
        by_key_methods += render_by_key_method(table_name, entity_name,
                                               col_items, unique_key)
    return f"""    async def insert_{entity_name}(
        self,
{kwargs}
    ):
        data = {{
{data_rows}
        }}
        return await self.save('{table_name}', data=data)
    
    async def select_{entity_name}(self, query, select_clause: str = '*', extra: dict = None):
        return await self.select_by_query('{table_name}', query, select_clause, extra)
    {by_key_methods}
"""


def render_by_key_method(table_name, entity_name, col_items, unique_key: list):
    indent = '    '
    key_name = '_'.join(unique_key)
    data_items = {
        key: value
        for key, value in col_items.items() if key not in unique_key
    }
    key_items = {
        key: value
        for key, value in col_items.items() if key in unique_key
    }
    kwargs = ''.join(f'{indent * 2}{name}: {py_type} = None,  # {comment}\n'
                     for name, (py_type, comment) in data_items.items())
    data_rows = ''.join(f"{indent * 3}'{name}': {name},\n"
                        for name in data_items.keys())
    args = ''.join(f'{indent * 2}{name}: {py_type},  # {comment}\n'
                   for name, (py_type, comment) in key_items.items())
    key_rows = ''.join(f"{indent * 3}'{name}': {name},\n"
                       for name in unique_key)
    return f"""
    async def update_{entity_name}_by_{key_name}(
        self,
{args}{kwargs}
    ):
        data = {{
{data_rows}
        }}
        key = {{
{key_rows}
        }}
        return await self.update_by_key('{table_name}', key=key, data=data)
    
    async def delete_{entity_name}_by_{key_name}(
        self,
{args}
    ):
        key = {{
{key_rows}
        }}
        return await self.delete_by_key('{table_name}', key=key)
    
    async def get_{entity_name}_by_{key_name}(
        self,
{args}
    ):
        key = {{
{key_rows}
        }}
        return await self.get_by_key('{table_name}', key=key)
"""



from lesscli import add_subcommand, run
from commondao.codegen import run_codegen


@add_subcommand('codegen', run_codegen)
def entry():
    """
    mysql service and toolkit for lessweb
    """
    pass


def main():
    run(entry)



from lesscli import add_argument
import os
from pathlib import Path
import re
from commondao.utils.templates import render_class_frame, render_class_methods
from commondao.utils.grammar import is_column_sql, parse_column_sql, guess_py_type, is_unique_key_sql, parse_unique_key_sql


def parse_create_table_sql(sql_content):
    """
    :return 形如table_name, col_items, unique_keys
       * col_items: dict[name, (py_type, comment)]
       * unique_keys: list[list[str]]
    """
    col_items = {}
    unique_keys = []
    seed = re.compile(r'CREATE TABLE *`?(\w+)`? *\((.*)\)',
                      flags=re.I | re.DOTALL)
    match_size = len(find_ret := seed.findall(sql_content))
    assert match_size == 1, 'cannot parse the create table SQL'
    table_name, sql_body = find_ret[0]
    for sql_line in sql_body.splitlines():
        if sql_line and is_column_sql(sql_line):
            col_name, sql_type, comment = parse_column_sql(sql_line)
            py_type = guess_py_type(sql_type)
            col_items[col_name] = py_type, comment
        elif sql_line and is_unique_key_sql(sql_line):
            col_names = parse_unique_key_sql(sql_line)
            unique_keys.append(col_names)
    return table_name, col_items, unique_keys


@add_argument('--schemadir',
              default='schema',
              help='schema direcotry, default: schema',
              required=False)
@add_argument('--output', help='mapper file for output', dest='outfile')
def run_codegen(schemadir, outfile):
    """
    Generate CommonDao source code
    """
    # list[table_name, entity_name, col_items, unique_keys]
    #   col_items: dict[name, (py_type, comment)]
    #   unique_keys: list[list[str]]
    class_body_blocks = []
    entity_filenames = os.listdir(schemadir)
    schema_path = Path(schemadir)
    for entity_filename in entity_filenames:
        entity_name = entity_filename.split('.', 1)[0]
        sql_content = (schema_path / entity_filename).open().read()
        table_name, col_items, unique_keys = parse_create_table_sql(
            sql_content)
        class_methods = render_class_methods(table_name, entity_name,
                                             col_items, unique_keys)
        class_body_blocks.append(class_methods)
    class_body = ''.join(class_body_blocks)
    class_frame = render_class_frame(class_body)
    with open(outfile, 'w') as fout:
        fout.write(class_frame)



from aiohttp.web import Application, Request
import aiomysql  # type: ignore
from aiomysql import Pool, Connection, Cursor, DictCursor
from lessweb.bridge import service
import logging
import re
from re import Match
from typing import TypeVar
from commondao.utils.common import script, and_, join, where

T = TypeVar('T')

SELECT_ONE = 1
SELECT_ALL = 2
EXECUTE = 3


@service
class Mysql:
    app: Application
    pool: Pool

    def __init__(self, app: Application):
        self.app = app


async def mysql_startup(app: Application, mysql: Mysql):
    mysql.pool = await aiomysql.create_pool(**app['config']['mysql'])


async def mysql_cleanup(mysql: Mysql):
    mysql.pool.close()
    await mysql.pool.wait_closed()


@service
class MysqlConn:
    request: Request
    mysql: Mysql
    conn: Connection
    cur: Cursor

    def __init__(self, request: Request, mysql: Mysql):
        self.request = request
        self.mysql = mysql


async def mysql_connect(handler, mysql_conn: MysqlConn):
    async with mysql_conn.mysql.pool.acquire() as conn:
        mysql_conn.conn = conn
        async with conn.cursor(DictCursor) as cur:
            mysql_conn.cur = cur
            return await handler()


class RegexCollect:
    words: list

    def __init__(self):
        self.words = []

    def repl(self, m: Match):
        word = m.group()
        self.words.append(word[2:])
        return word[0] + '%s'

    def build(self, sql: str, params: dict) -> tuple:
        pattern = r"[^:]:[a-zA-Z][\w.]*"
        pg_sql = re.sub(pattern, self.repl, sql)
        pg_params = tuple(params[k] for k in self.words)
        return pg_sql, pg_params


def make_sub_clause(query) -> tuple:
    """
    query: Dict | MutableMapping
    return: (value_map, limit_map, where_clause, order_clause, limit_clause)
    """
    eq_filter = '<field>=:<field>.value'
    filter_map = {}  # filter_map[query_key] = filter
    limit_map = {'limit': 1000}
    orderby_rules = []
    value_map = {}
    for query_key, query_val in query.items():
        if query_key == 'limit':
            limit_map['limit'] = int(query_val)
        elif query_key == 'offset':
            limit_map['offset'] = int(
                query_val
            )  # todo 如果传入的limit=0，则只查询总数；否则正常分页查询。如果不传limit则默认1000。
        elif '.' not in query_key:
            filter_map.setdefault(query_key, eq_filter.replace('<field>', query_key))
            value_map[f'{query_key}.value'] = query_val
        elif query_key.endswith('.value'):
            filter_map.setdefault(query_key[:-6], eq_filter.replace('<field>', query_key[:-6]))
            value_map[query_key] = query_val
        elif query_key.endswith('.filter'):
            filter_map[query_key[:-7]] = query_val.replace('<field>', query_key[:-7])
        elif query_key.endswith('.order'):
            assert query_val in 'asc' or query_val in 'desc'
            orderby_rules.append(query_key[:-6] + ' ' + query_val)
        else:
            value_map[query_key] = query_val
    where_clause = ''
    order_clause = ', '.join(orderby_rules)
    limit_clause = 'limit %d' % int(limit_map['limit'])
    for filter_val in filter_map.values():
        where_clause += ('WHERE' if not where_clause else 'AND')\
                        + f' ({filter_val}) '
    if 'offset' in limit_map:
        limit_clause += ' offset %d' % int(limit_map['offset'])
    return value_map, limit_map, where_clause, order_clause, limit_clause


@service
class Mapper:
    conn: MysqlConn

    def __init__(self, conn: MysqlConn):
        self.conn = conn

    async def commit(self):
        await self.conn.conn.commit()

    async def execute(self, mode, sql: str, data: dict):
        cursor = self.conn.cur
        logging.info(sql)
        pg_sql, pg_params = RegexCollect().build(sql, data)
        logging.info('%s => %s', pg_sql, pg_params)
        await cursor.execute(pg_sql, pg_params)
        if mode == SELECT_ONE:
            return await cursor.fetchone() or {}
        elif mode == SELECT_ALL:
            return await cursor.fetchall()
        else:
            return cursor.rowcount

    async def select_one(self, sql: str, data: dict):
        return await self.execute(SELECT_ONE, sql, data)

    async def select_all(self, sql: str, data: dict):
        return await self.execute(SELECT_ALL, sql, data)

    async def insert(self, sql: str, data: dict):
        return await self.execute(EXECUTE, sql, data)

    async def update(self, sql: str, data: dict):
        return await self.execute(EXECUTE, sql, data)

    async def delete(self, sql: str, data: dict):
        return await self.execute(EXECUTE, sql, data)

    async def save(self, tablename: str, *, data: dict):
        selected_data = {
            key: value
            for key, value in data.items() if value is not None
        }
        sql = script(
            'insert into',
            tablename,
            '(',
            join(*selected_data.keys()),
            ') values (',
            join(*[f':{key}' for key in selected_data.keys()]),
            ')',
        )
        return await self.insert(sql, selected_data)

    async def update_by_key(self, tablename, *, key: dict, data: dict):
        selected_data = {
            key: value
            for key, value in data.items() if value is not None
        }
        if not selected_data:
            return 0
        sql = script(
            'update',
            tablename,
            'set',
            join(*[f'{k}=:{k}' for k in selected_data.keys()], ),
            'where',
            and_(*[f'{k}=:{k}' for k in key.keys()]),
        )
        return await self.update(sql, {**data, **key})

    async def delete_by_key(self, tablename, *, key: dict):
        sql = script(
            'delete from',
            tablename,
            'where',
            and_(*[f'{k}=:{k}' for k in key.keys()]),
        )
        return await self.delete(sql, key)

    async def get_by_key(self, tablename, *, key: dict):
        sql = script('select * from', tablename,
                     where(and_(*[f'{k}=:{k}' for k in key.keys()])),
                     'limit 1')
        return await self.select_one(sql, key)

    async def select_by_query(self,
                              tablename,
                              query,
                              select_clause: str = '*',
                              extra: dict = None):
        """
        输入：
            query: Dict | MutableMapping
            extra: Dict
        保留字：limit, offset, .value, .field, .order, <field>
        <field>.value: 可以省略.value
        <field>.filter: 语法例如"<field> between <value> and <end>"或者"<filter> >= <value>"
        <field>.order: 值可取"asc|desc"(不支持大写)
        tablename分为表名模式和完整模式，完整模式包含最多一个<select>, <where>, <order>, <limit>
        如果传入的limit=0，则只查询总数；如果不传limit或limit is None，则不分页；否则正常分页查询。
        """
        value_map, limit_map, where_clause, order_clause, limit_clause = make_sub_clause(
            query)
        count_sql = script(
            'SELECT COUNT(*) AS total FROM',
            tablename,
            where_clause,
        )
        if extra:
            value_map.update(extra)
        count_ret = await self.select_one(count_sql, value_map)
        assert select_clause == '*' or re.match(
            r'[\w,` ]+',
            select_clause), f'Invalid select_clause: {select_clause}'
        records_sql = script(
            'SELECT',
            select_clause,
            'FROM',
            tablename,
            where_clause,
            f'ORDER BY {order_clause}' if order_clause else '',
            limit_clause,
        )
        records = await self.select_all(records_sql, value_map)
        return {**count_ret, **limit_map, 'list': records}



import os


def project_to_package(project):
    """
    >>> print(project_to_package('abcdef'))
    abcdef
    >>> print(project_to_package('abc_DEF'))
    abc_DEF
    >>> print(project_to_package('abc-D..EF'))
    abc_D_EF
    """
    tmp = []
    for c in project:
        if 'a' <= c.lower() <= 'z' or c == '_':
            tmp.append(c)
        elif not tmp or tmp[-1] != '_':
            tmp.append('_')
        else:
            pass
    return ''.join(tmp)


def captitalize(s):
    if not s:
        return s
    return s[0].upper() + s[1:]


def camelize(package, upper=True):
    """
    >>> print(camelize('abcdef', False))
    abcdef
    >>> print(camelize('abc_DEF'))
    AbcDEF
    >>> print(camelize('abc_d_efg', False))
    abcDEfg
    """
    tmp = []
    package = package.replace('-', '_')
    for w in package.split('_'):
        tmp.append(captitalize(w))
    if not upper and tmp:
        tmp[0] = tmp[0].lower()
    return ''.join(tmp)


def exe(cmd):
    print(cmd)
    ret = os.system(cmd)
    return ret == 0


def not_ready(cmd):
    return os.system("which " + cmd) != 0


def assert_or_exit(expect, message='failed!'):
    if not expect:
        print(message)
        exit(1)


def dump_file_name(filename: str, project: str):
    if '.' in filename[1:]:
        prefix, suffix = filename.rsplit('.', 1)
        return f'{prefix}.{project}.{suffix}'
    else:
        return f'{filename}.{project}'


def script(*segs) -> str:
    return ' '.join(seg or '' for seg in segs)


def join(*segs) -> str:
    return ','.join(seg or '' for seg in segs)


def and_(*segs) -> str:
    ret = ' and '.join(seg or '' for seg in segs).strip()
    return f'({ret})' if ret else ''


def where(*segs) -> str:
    sql = and_(*segs)
    return f'where {sql} ' if sql else ''



import ast
import re


def is_unique_key_sql(sql_line: str):
    return sql_line.strip().upper().startswith(
        ('PRIMARY ', 'UNIQUE ')) and parse_unique_key_sql(sql_line)


def parse_unique_key_sql(sql_line: str):
    """
    返回：list[col_name]
    例子：
        PRIMARY KEY (`id`),
        UNIQUE KEY `uni_phone` (`phone`, `openid`)
    """
    return re.findall(r'\w+', sql_line.split('(', 1)[-1].split(')', 1)[0])


def is_column_sql(sql_line: str):
    """
    判断一个SQL行是否是合法的列定义SQL
    """
    return not sql_line.strip().upper().startswith(
        ('PRIMARY ', 'KEY ', 'CONSTRAINT ', 'UNIQUE '))


def parse_column_sql(sql_line: str):
    """
    返回：col_name, sql_type, comment
    """
    sql_line = sql_line.strip().rstrip(',')
    if ' COMMENT ' in sql_line:
        sql_line, comment_part = sql_line.split(' COMMENT ', 1)
        comment = ast.literal_eval(comment_part)
    else:
        comment = ''
    col_name, sql_type = sql_line.split(None, 2)[:2]
    col_name = col_name.strip('`')
    return col_name, sql_type, comment


def guess_py_type(sql_type):
    sql_type = sql_type.lower()
    if sql_type == 'date':
        return 'date'
    elif sql_type == 'datetime':
        return 'datetime'
    elif sql_type == 'json':
        return 'str'
    elif sql_type.startswith('bigint'):
        return 'int'
    elif sql_type.startswith('int'):
        return 'int'
    elif sql_type.startswith('enum'):
        return 'str'  # 后续改进
    elif sql_type.startswith('tinyint'):
        return 'bool'
    else:
        return 'str'



from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
import re
import ast

_version_re = re.compile(r'__version__\s+=\s+(.*)')
version = str(ast.literal_eval(
    _version_re.search(
        open('commondao/__init__.py').read()
    ).group(1)
))


setup(
        name='lessweb-commondao',
        version=version,
        description='mysql service and toolkit for lessweb',
        long_description='\nREADME: https://github.com/lessweb/lessweb-commondao\n\n'
                         'Cookbook: http://www.lessweb.cn',
        url='https://github.com/lessweb/lessweb-commondao',
        author='qorzj',
        author_email='goodhorsezxj@gmail.com',
        license='Apache 2',
        platforms=['any'],

        classifiers=[
            ],
        keywords='lessweb mysql',
        packages=['commondao', 'commondao.utils'],
        install_requires=['aiohttp', 'lesscli', 'lessweb', 'mysql-connector', 'aiomysql'],
        entry_points={
            'console_scripts': [
                'commondao = commondao.index:main',
                ],
            },
    )
